# -*- coding: utf-8 -*-
#
# FinLogic - Brazilian Company Financial Data
# https://github.com/crdcj/FinLogic
#
# Copyright 2022 Carlos Carvalho
#
from .data import load, search_company, info, search_segment, rank
from .company import Company

__version__ = "0.6.1"
__author__ = "Carlos Carvalho"

__all__ = ["Company", "load", "search_company", "info", "search_segment", "rank"]



"""This module contains the Company class, which provides methods for obtaining
financial reports and calculating financial indicators for a company. Users can
set the accounting method, unit, tax rate, and language for the company object.

Classes:
    Company: Represents a company with its financial information and allows
    users to generate financial reports and indicators.

Main variables:
    cfg.DF  memory dataframe with all accounting data
    _df     company dataframe (protected)
    dfi     function input dataframe
    dfo     function output dataframe

RuntimeWarning:
    Pandas query + numexpr module -> Engine has switched to 'python' because
    numexpr does not support extension array dtypes (category not supported
    yet). Please set your engine to python manually. That means that the query
    method has to use engine='python' to work with category dtype. N.B. Only
    FinLogic Dataframe uses category dtype for efficiency.

"""
from typing import Literal
import pandas as pd
from . import data as dt
from . import indicators as ic


class Company:
    """A class to represent a company financial data.

     This class provides methods to create financial reports and to calculate
     financial indicators based on a company's accounting data. The class also
     has an AI generated dictionary to translate from Portuguese to English.

    Attributes:
         identifier: A unique identifier for the company. Both CVM ID (int) and
            Fiscal ID (str) can be used.
         is_consolidated: The accounting methods can be either 'con' for consolidated or
            'sep' for separate. Defaults to 'con' (str).
         acc_unit: The accounting unit for the financial statements where "t"
            represents thousands, "m" represents millions and "b" represents
            billions (int, float or str). Defaults to 1.
         tax_rate: The tax rate for the company. Defaults to 0.34, which is
            the standard corporate tax rate in Brazil (float).
         language: The language for the financial reports. Options are "english"
            or "portuguese". Defaults to "english" (str).

     Methods:
         report: Creates a financial report for the company.
         custom_report: Creates a custom financial report for the company.
         indicators: Calculates the financial indicators of the company.

     Raises:
         ValueError: If the input arguments are invalid.
    """

    def __init__(
        self,
        identifier: int | str,
        is_consolidated: bool = True,
        acc_unit: float | Literal["t", "m", "b"] = 1.0,
        tax_rate: float = 0.34,
        language: Literal["english", "portuguese"] = "english",
    ):
        """Initializes a new instance of the Company class."""
        self._initialized = False
        self.identifier = identifier
        self.is_consolidated = is_consolidated
        self.acc_unit = acc_unit
        self.tax_rate = tax_rate
        self.language = language
        self._initialized = True
        # Only set _df after identifier, is_consolidated and acc_unit are setted
        self._set_df()

    @staticmethod
    def convert_to_sl(expr: str) -> str:
        """Converts a string to a single line."""
        return expr.replace("\n", "")

    @property
    def identifier(self) -> int | str:
        """Set a unique identifier to select the company in FinLogic Database.

        This method sets the company's CVM and fiscal ID based on a given
        identifier. The identifier can be either the CVM ID or the Fiscal ID
        (CNPJ). If the identifier is not found in the database, a KeyError is
        raised.

        Args:
            identifier: A unique identifier to select a company in FinLogic
                Database. Both CVM ID or Fiscal ID can be used. CVM ID
                (regulator number) must be an integer. Fiscal ID must be a
                string in 'XX.XXX.XXX/XXXX-XX' format.

        Returns:
            None

        Raises:
            KeyError: If the given identifier isn't found in Finlogic Database.
        """
        return self._identifier

    @identifier.setter
    def identifier(self, identifier: int | str):
        # Create custom data frame for ID selection
        df = (
            dt.FINANCIALS_DF[["cvm_id", "tax_id", "name_id"]]
            .query("cvm_id == @identifier or tax_id == @identifier")
            .drop_duplicates(ignore_index=True)
        )
        if not df.empty:
            self._cvm_id = df.loc[0, "cvm_id"]
            self.tax_id = df.loc[0, "tax_id"]
            self.name_id = df.loc[0, "name_id"]
            self._identifier = identifier
        else:
            raise KeyError(f"Company 'identifier' {identifier} not found.")
        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_df()

    @property
    def is_consolidated(self) -> bool:
        """Gets or sets the accounting method for registering investments in
        subsidiaries.

        The "is_consolidated" must be True for consolidated or False for separate.
        Consolidated accounting combines the financial statements of a parent
        company and its subsidiaries, while separate accounting keeps them
        separate. Defaults to 'consolidated'.

        Raises:
            ValueError: If the accounting method is invalid.
        """
        return self._is_consolidated

    @is_consolidated.setter
    def is_consolidated(self, value: bool):
        if type(value) is bool:
            self._is_consolidated = value
        else:
            raise ValueError("Company 'is_consolidated' value is invalid")
        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_df()

    @property
    def acc_unit(self) -> float:
        """Gets or sets the accounting unit for the financial statements.

        The "acc_unit" is a constant that will divide all company
        accounting values. The constant must be a number greater than
        zero or one of the following strings:
            - "t" to represent thousands       (1,000)
            - "m" to represent millions     (1,000,000)
            - "b" to represent billions (1,000,000,000)

        Returns:
            The current accounting unit.

        Raises:
            ValueError: If the accounting unit is invalid.

        Examples:
            To set the accounting unit to millions:
                company.acc_unit = "m"

            To set the accounting unit to a custom factor, e.g., 10,000:
                company.acc_unit = 10_000
        """
        return self._acc_unit

    @acc_unit.setter
    def acc_unit(self, value: float | Literal["t", "m", "b"]):
        match value:
            case "t":
                self._acc_unit = 1_000.0
            case "m":
                self._acc_unit = 1_000_000.0
            case "b":
                self._acc_unit = 1_000_000_000.0
            case str():  # Add this case to catch invalid strings
                raise ValueError("Invalid string for Accounting Unit")
            case v if v > 0:
                self._acc_unit = float(v)
            case _:
                raise ValueError("Accounting Unit is invalid")

        # If object was already initialized, reset company dataframe
        if self._initialized:
            self._set_df()

    @property
    def tax_rate(self) -> float:
        """Gets or sets company 'tax_rate' property.

        The "tax_rate" is used to calculate some of the company
        indicators, such as EBIT and net income. The default value
        is 0.34, which is the standard corporate tax rate in Brazil.

        Returns:
            The tax rate value.

        Raises:
            ValueError: If the tax rate is not a float between 0 and 1.

        Examples:
            To set the tax rate to 21%:
                company.tax_rate = 0.21

            To set the tax rate to 0% (tax exempt):
                company.tax_rate = 0.0
        """
        return self._tax_rate

    @tax_rate.setter
    def tax_rate(self, value: float):
        if 0 <= value <= 1:
            self._tax_rate = value
        else:
            raise ValueError("Company 'tax_rate' value is invalid")

    @property
    def language(self) -> str:
        """Gets or sets the language of the account names.

        It is the language used in the account names of the financial
        statements. This property accepts a string representing the
        desired language. Supported languages are "english" and
        "portuguese". The default is 'english'.

        Returns:
            The language used in the account names.

        Raises:
            KeyError: If the provided language is not supported.
        """

        return self._language

    @language.setter
    def language(self, language: Literal["english", "portuguese"]):
        # Supported languages
        list_languages = ["english", "portuguese"]
        if language.lower() in list_languages:
            self._language = language.capitalize()
        else:
            sup_lang = f"Supported languages: {', '.join(list_languages)}"
            raise KeyError(f"'{language}' not supported. {sup_lang}")

    def _set_df(self) -> pd.DataFrame:
        """Sets the company data frame.

        This method creates a dataframe with the company's financial
        statements.
        """
        expr = "cvm_id == @self._cvm_id and is_consolidated == @self._is_consolidated"
        df = dt.FINANCIALS_DF.query(expr).reset_index(drop=True)

        # Convert category columns back to string
        columns = df.columns
        cat_cols = [c for c in columns if df[c].dtype == "category"]
        df[cat_cols] = df[cat_cols].astype("string")

        # Adjust for unit change only where it is not EPS (acc_code 8...)
        mask = ~df["acc_code"].str.startswith("3.99")
        df.loc[mask, "acc_value"] = df.loc[mask, "acc_value"] / self._acc_unit

        self._first_period = df["period_end"].min()
        self._last_period = df["period_end"].max()

        # Not necessarily there will be a quarterly report for the last period
        self._last_annual = df.query("is_annual")["period_end"].max()

        if self._last_period == self._last_annual:
            self._last_period_type = "annual"
            self._last_quarterly = None
        else:
            self._last_period_type = "quarterly"
            self._last_quarterly = df.query("not is_annual")["period_end"].max()

        # Drop columns that are already company attributes or will not be used
        df.drop(
            columns=["name_id", "cvm_id", "tax_id", "is_consolidated"],
            inplace=True,
        )

        # Set company data frame
        self._df = df

    def info(self) -> pd.DataFrame:
        """Print a concise summary of a company."""
        if self._df.empty:
            acc_method = "consolidated" if self._is_consolidated else "separate"
            print("There is no avaible accounting data for:")
            print(f"    cvm_id = {self._cvm_id} and accounting method = {acc_method}")
            return None
        company_info = {
            "Name": self.name_id,
            "CVM ID": self._cvm_id,
            "Fiscal ID (CNPJ)": self.tax_id,
            "Total Accounting Rows": len(self._df.index),
            "Selected Accounting Method": "consolidated"
            if self._is_consolidated
            else "separate",
            "Selected Accounting Unit": self._acc_unit,
            "Selected Tax Rate": self._tax_rate,
            "First Report": self._first_period.strftime("%Y-%m-%d"),
            "Last Report": self._last_period.strftime("%Y-%m-%d"),
        }
        s = pd.Series(company_info)
        s.name = "Company Info"
        return s.to_frame()

    @staticmethod
    def _build_report_index(dfi: pd.DataFrame) -> pd.DataFrame:
        """Build the index for the report. This function is used by the
        _build_report function. The index is built from the annual reports
        "acc_code" works as a primary key. Other columns set the preference order
        """
        df = (
            dfi[["acc_code", "acc_name", "period_end"]]
            .sort_values(by=["acc_code", "period_end"])
            .drop_duplicates(subset=["acc_code"], keep="last", ignore_index=True)[
                ["acc_code", "acc_name"]
            ]
        )
        return df

    def _build_report(self, dfi: pd.DataFrame) -> pd.DataFrame:
        # Start "dfo" with the index
        dfo = self._build_report_index(dfi)
        year_cols = ["acc_code", "acc_value"]
        periods = sorted(dfi["period_end"].drop_duplicates())
        for period in periods:
            df_year = dfi.query("period_end == @period")[year_cols].copy()
            period_str = period.strftime("%Y-%m-%d")
            if period == self._last_period and self._last_period_type == "quarterly":
                period_str += " ltm"
            df_year.rename(columns={"acc_value": period_str}, inplace=True)
            dfo = pd.merge(dfo, df_year, how="left", on=["acc_code"])
        dfo.fillna(0, inplace=True)
        return dfo.sort_values("acc_code", ignore_index=True)

    @staticmethod
    def _remove_not_last_quarters(df: pd.DataFrame) -> pd.DataFrame:
        """Remove quarters that are not the last one.

        This function removes quarters that are not the last one.
        This is useful when generating reports.

        Args:
            df: Dataframe with the financial statements.

        Returns:
            Dataframe with the financial statements without the quarters that are not
            the last one.
        """
        mask1 = ~df["is_annual"]
        mask2 = df["period_end"] != df["period_end"].max()
        mask = mask1 & mask2
        df = df[~mask].reset_index(drop=True)
        return df

    def report(
        self,
        report_type: Literal[
            "balance_sheet",
            "assets",
            "cash",
            "current_assets",
            "non_current_assets",
            "liabilities",
            "debt",
            "current_liabilities",
            "non_current_liabilities",
            "liabilities_and_equity",
            "equity",
            "income_statement",
            "cash_flow",
            "earnings_per_share",
        ],
        acc_level: Literal[0, 1, 2, 3, 4] = 0,
        num_years: int = 0,
    ) -> pd.DataFrame:
        """Generate an accounting report for the company.

        This function generates a report representing one of the financial
        statements for the company adjusted by the attributes passed.

        Args:
            report_type: Type of financial report to be generated. Options are:
                - balance_sheet
                    - assets
                        - cash
                        - current_assets,
                        - non_current_assets
                    - liabilities
                        - debt
                        - current_liabilities
                        - non_current_liabilities,
                    - liabilities_and_equity
                        - equity
                - income_statement
                - cash_flow
                - earnings_per_share
                - comprehensive_income,
                - added_value
            acc_level: Detail level to show for account codes. Options are 0, 1,
                2, 3 or 4. Defaults to 0. How the values works:
                    0    -> X...       (show all accounts)
                    1    -> X          (show 1 level)
                    2    -> X.YY       (show 2 levels)
                    3    -> X.YY.ZZ    (show 3 levels)
                    4    -> X.YY.ZZ.WW (show 4 levels)
            num_years: Number of years to include in the report. Defaults to 0
                (all years).

        Returns:
            pd.DataFrame: Generated financial report as a pandas DataFrame.

        Raises:
            ValueError: If some argument is invalid.
        """
        # Copy company dataframe to avoid changing it
        df = self._df.copy()
        df = self._remove_not_last_quarters(df)
        # Check input arguments.
        if acc_level not in [0, 1, 2, 3, 4]:
            raise ValueError("acc_level expects 0, 1, 2, 3 or 4")

        # Filter dataframe for selected acc_level
        # Example of an acc_code: "7.08.04.04" -> 4 levels and 3 dots
        if acc_level:
            df.query(rf"acc_code.str.count('\.') <= {acc_level - 1}", inplace=True)

        # Set language
        class MyDict(dict):
            """Custom dictionary class to return key if key is not found."""

            def __missing__(self, key):
                return "(pt) " + key

        if self._language == "English":
            _pten_dict = dict(dt.LANGUAGE_DF.values)
            _pten_dict = MyDict(_pten_dict)
            df["acc_name"] = df["acc_name"].map(_pten_dict)

        """
        Filter dataframe for selected acc_code
        df['acc_code'].str[0].unique() -> [1, 2, 3, 4, 5, 6, 7]
        The first part of 'acc_code' is the report type
        Table of reports correspondence:
            1 -> Balance Sheet - Assets
            2 -> Balance Sheet - Liabilities and Shareholders’ Equity
            3 -> Income
            4 -> Comprehensive Income
            5 -> Changes in Equity
            6 -> Cash Flow (Indirect Method)
            7 -> Added Value
            8 -> Earnings per Share
        """
        report_types = {
            "balance_sheet": ("1", "2"),
            "assets": ("1"),
            "cash": ("1.01.01", "1.01.02"),
            "current_assets": ("1.01"),
            "non_current_assets": ("1.02"),
            "liabilities": ("2.01", "2.02"),
            "debt": ("2.01.04", "2.02.01"),
            "current_liabilities": ("2.01"),
            "non_current_liabilities": ("2.02"),
            "liabilities_and_equity": ("2"),
            "equity": ("2.03"),
            "income_statement": ("3"),
            "earnings_per_share": ("3.99"),
            "cash_flow": ("6"),
        }
        acc_codes = report_types[report_type]  # noqa
        df.query("acc_code.str.startswith(@acc_codes)", inplace=True)
        df.reset_index(drop=True, inplace=True)

        # Show only selected years
        all_periods = sorted(df["period_end"].drop_duplicates())
        selected_periods = all_periods[-num_years:]  # noqa
        df.query("period_end in @selected_periods", inplace=True)

        return self._build_report(df)

    def custom_report(
        self,
        acc_list: list[str],
        num_years: int = 0,
    ) -> pd.DataFrame:
        """Generate a financial report from custom list of accounting codes

        Args:
            acc_list: A list of strings containg accounting codes to be used
                in the report
            num_years: Select how many years to show in the report.
                Defaults to 0 (show all years)

        Returns:
            pd.DataFrame: Generated financial report as a pandas DataFrame.

        Raises:
            ValueError: If some argument is invalid.
        """
        df_bs = self.report("balance_sheet", num_years=num_years)
        df_is = self.report("income_statement", num_years=num_years)
        df_cf = self.report("cash_flow", num_years=num_years)
        df = (
            pd.concat([df_bs, df_is, df_cf])
            .query(f"acc_code == {acc_list}")
            .reset_index(drop=True)
        )
        return df

    def indicators(self, num_years: int = 0) -> pd.DataFrame:
        """Calculate the company main operating indicators.

        Args:
            num_years: Number of years to consider for calculation. If 0, use
                all available years. Defaults to 0.

        Returns:
            pd.DataFrame: Dataframe containing calculated financial indicators.
        """
        expr = "cvm_id == @self._cvm_id and is_consolidated == @self._is_consolidated"
        df = dt.INDICATORS_DF.query(expr)
        df = ic.format_indicators(df, unit=self._acc_unit)
        # Columns cvm_id and is_consolidated are redundant for the Company class
        df.drop(columns=["cvm_id", "is_consolidated"], inplace=True)
        # Show only the selected number of years
        if num_years > 0:
            df = df[df.columns[-num_years:]].copy()

        return df



import finlogic as fl
import pytest

fl.load()


def test_identifier():
    """Test the identifier method of the Company class."""
    petro_sep = fl.Company(9512, is_consolidated=False, acc_unit="b")

    test_cvm_id = 4170
    test_tax_id = "33.592.510/0001-54"
    petro_sep.identifier = test_cvm_id
    assert petro_sep._cvm_id == test_cvm_id
    assert petro_sep.tax_id == test_tax_id

    petro_sep.identifier = test_tax_id
    assert petro_sep._cvm_id == test_cvm_id
    assert petro_sep.tax_id == test_tax_id

    invalid_identifier = 99999999  # Use an invalid identifier here

    with pytest.raises(KeyError):
        petro_sep.identifier = invalid_identifier


def test_attributes():
    """Test the info method of the Company class."""
    petro_sep = fl.Company(9512, is_consolidated=False, acc_unit="b")

    # Check the attributes
    assert petro_sep.name_id == "PETROLEO BRASILEIRO S.A. PETROBRAS"
    assert petro_sep._cvm_id == 9512
    assert petro_sep.tax_id == "33.000.167/0001-01"


def test_report():
    """Test if the report method of the Company class returns the correct values."""
    petro_con = fl.Company(9512, is_consolidated=True, acc_unit="b")

    petro_report = petro_con.report(report_type="assets")
    # Get Total Assets
    assets_2009 = round(petro_report.query("acc_code == '1'")["2009-12-31"][0], 3)
    assets_2015 = round(petro_report.query("acc_code == '1'")["2015-12-31"][0], 3)
    assets_2020 = round(petro_report.query("acc_code == '1'")["2020-12-31"][0], 3)
    # Check the reports
    assert assets_2009 == 350.419
    assert assets_2015 == 900.135
    assert assets_2020 == 987.419


def test_indicators():
    """Test the indicators method of the Company class."""
    petro_sep = fl.Company(9512, is_consolidated=False, acc_unit="b")
    petro_con = fl.Company(9512, is_consolidated=True, acc_unit="b")
    # Petro indicators (separate)
    petro_indicators_sep = petro_sep.indicators()
    petro_indicators_con = petro_con.indicators()
    # Get the indicators (rounded to 4 decimals)
    revenues_2009_sep = round(petro_indicators_sep.at["revenues", "2009-12-31"], 4)
    total_debt_2015_sep = round(petro_indicators_sep.at["total_debt", "2015-12-31"], 4)
    roic_2021_sep = round(petro_indicators_sep.at["roic", "2021-12-31"], 4)
    # Get the indicators (rounded to 4 decimals)
    revenues_2009_con = round(petro_indicators_con.at["revenues", "2009-12-31"], 4)
    total_debt_2015_con = round(petro_indicators_con.at["total_debt", "2015-12-31"], 4)
    roic_2021_con = round(petro_indicators_con.at["roic", "2021-12-31"], 4)
    # Check the indicators
    assert roic_2021_sep == 0.1512
    assert revenues_2009_sep == 134.0339
    assert total_debt_2015_sep == 305.3460
    # Check the indicators
    assert roic_2021_con == 0.2149
    assert revenues_2009_con == 182.8338
    assert total_debt_2015_con == 493.0230



from datetime import date
import finlogic as fl
from finlogic import data

fl.load()


def test_info():
    """Test the info method of the Database module."""
    info = data.info()["FinLogic Info"]
    first_report = date.fromisoformat(info["first_report"])
    assert first_report.year == 2009
    assert info["accounting_entries"] > 100_000


def test_search_company():
    """Test the search_company method of the Database module."""
    search_result = fl.search_company("3r")
    """
        name_id cvm_id  tax_id  segment is_restructuring    most_traded_stock
    0	3R ...  25291   12...   expl... False	            RRRP3
    """
    # Check results
    assert set(search_result["cvm_id"]) == {25291}



import pandas as pd


TAX_RATE = 0.34
INDICATORS_CODES = {
    "1": "total_assets",
    "1.01": "current_assets",
    "1.01.01": "cash_equivalents",
    "1.01.02": "financial_investments",
    "2.01": "current_liabilities",
    "2.01.04": "short_term_debt",
    "2.02.01": "long_term_debt",
    "2.03": "equity",
    "3.01": "revenues",
    "3.03": "gross_profit",
    "3.05": "ebit",
    "3.07": "ebt",
    "3.08": "effective_tax",
    "3.11": "net_income",
    "6.01": "operating_cash_flow",
    "6.01.01.04": "depreciation_amortization",
    "3.99.01.01": "eps",
}


def filter_indicators_data(dfi: pd.DataFrame) -> pd.DataFrame:
    codes = list(INDICATORS_CODES.keys())  # noqa: used in query below
    """There are 137 repeated entries in 208784 rows. These are from companies
    with some exotic period_end dates, as for cvm_id 3450. These entries will be
    removed in the next step, when we drop duplicates and the last entry
    published will be kept.
    """
    drop_cols = ["tax_id", "acc_name", "period_begin"]
    sort_cols = ["cvm_id", "is_consolidated", "acc_code", "period_end"]
    subset_cols = ["cvm_id", "is_consolidated", "acc_code", "period_end"]

    dfo = (
        dfi.query("acc_code in @codes")
        .drop(columns=drop_cols)
        # .query("cvm_id == 9512 and is_consolidated")  # for testing
        .sort_values(by=sort_cols, ignore_index=True)
        .drop_duplicates(subset=subset_cols, keep="last", ignore_index=True)
        .astype({"acc_code": "string"})
    )
    return dfo


def pivot_df(df) -> pd.DataFrame:
    index_cols = ["cvm_id", "name_id", "is_annual", "is_consolidated", "period_end"]
    dfp = (
        pd.pivot(df, values="acc_value", index=index_cols, columns=["acc_code"])
        .fillna(0)
        .reset_index()
    )
    return dfp


def insert_annual_avg_col(col_name: str, df: pd.DataFrame) -> pd.DataFrame:
    gp_cols = ["cvm_id", "is_annual", "is_consolidated"]
    col_name_p = f"{col_name}_p"
    avg_col_name = f"avg_{col_name}"
    df[col_name_p] = df.groupby(by=gp_cols)[col_name].shift(1)
    df[col_name_p] = df[col_name_p]
    df[col_name_p].fillna(df[col_name], inplace=True)
    df[avg_col_name] = (df[col_name] + df[col_name_p]) / 2
    df.drop(columns=[col_name_p], inplace=True)
    return df


def insert_quarterly_avg_col(col_name: str, df: pd.DataFrame) -> pd.DataFrame:
    gp_cols = ["cvm_id", "is_annual", "is_consolidated"]
    col_name_p4 = f"{col_name}_p4"
    col_name_p1 = f"{col_name}_p1"
    col_name_p = f"{col_name}_p"
    avg_col_name = f"avg_{col_name}"
    df[col_name_p4] = df.groupby(by=gp_cols)[col_name].shift(4)
    df[col_name_p1] = df.groupby(by=gp_cols)[col_name].shift(1)
    df[col_name_p] = df[col_name_p4]
    df[col_name_p].fillna(df[col_name_p1], inplace=True)
    df[col_name_p].fillna(df[col_name], inplace=True)
    df[avg_col_name] = (df[col_name] + df[col_name_p]) / 2
    df.drop(columns=[col_name_p4, col_name_p1, col_name_p], inplace=True)
    return df


def insert_key_cols(df: pd.DataFrame) -> pd.DataFrame:
    df["total_cash"] = df["cash_equivalents"] + df["financial_investments"]
    df.drop(columns=["cash_equivalents", "financial_investments"], inplace=True)

    df["total_debt"] = df["short_term_debt"] + df["long_term_debt"]
    df["net_debt"] = df["total_debt"] - df["total_cash"]
    df.drop(columns=["short_term_debt", "long_term_debt"], inplace=True)

    df["working_capital"] = df["current_assets"] - df["current_liabilities"]
    df["effective_tax_rate"] = -1 * df["effective_tax"] / df["ebt"]
    df["ebitda"] = df["ebit"] + df["depreciation_amortization"]
    df["invested_capital"] = df["total_debt"] + df["equity"] - df["total_cash"]

    return df


def process_indicators(df, is_annual: bool, insert_avg_col) -> pd.DataFrame:
    df.rename(columns=INDICATORS_CODES, inplace=True)
    df = insert_key_cols(df)

    avg_cols = ["invested_capital", "total_assets", "equity"]
    for col_name in avg_cols:
        df = insert_avg_col(col_name, df)

    # For quarterly data, we need only the last row of each group
    if not is_annual:
        gp_cols = ["cvm_id", "is_annual", "is_consolidated"]
        df = df.groupby(by=gp_cols).tail(1).dropna().reset_index(drop=True)

    # Margin ratios
    CUT_OFF_VALUE = 1_000_000
    df["gross_margin"] = df["gross_profit"] / df["revenues"]
    df["ebitda_margin"] = df["ebitda"] / df["revenues"]
    df["operating_margin"] = df["ebit"] / df["revenues"]
    df["net_margin"] = df["net_income"] / df["revenues"]
    margin_cols = ["gross_margin", "ebitda_margin", "operating_margin", "net_margin"]
    df.loc[df["revenues"] <= CUT_OFF_VALUE, margin_cols] = 0

    # Return ratios
    df["return_on_assets"] = df["ebit"] * (1 - TAX_RATE) / df["avg_total_assets"]
    df.loc[df["avg_total_assets"] <= CUT_OFF_VALUE, "return_on_assets"] = 0
    df["return_on_equity"] = df["ebit"] * (1 - TAX_RATE) / df["avg_equity"]
    df.loc[df["avg_equity"] <= CUT_OFF_VALUE, "return_on_equity"] = 0
    df["roic"] = df["ebit"] * (1 - TAX_RATE) / df["avg_invested_capital"]
    df.loc[df["avg_invested_capital"] <= CUT_OFF_VALUE, "roic"] = 0

    # Drop avg_cols
    avg_cols = ["avg_total_assets", "avg_equity", "avg_invested_capital"]
    df.drop(columns=avg_cols, inplace=True)

    return df


def build_indicators(financials_df: pd.DataFrame) -> pd.DataFrame:
    """Build indicators dataframe."""
    start_df = filter_indicators_data(financials_df)

    # Construct pivot tables for annual and quarterly
    dfa = pivot_df(start_df.query("is_annual"))
    dfq = pivot_df(start_df.query("not is_annual"))

    # Build indicators
    dfai = process_indicators(dfa, True, insert_annual_avg_col)
    dfqi = process_indicators(dfq, False, insert_quarterly_avg_col)

    # Build output dataframe
    sort_cols = ["cvm_id", "is_consolidated", "period_end"]
    df = pd.concat([dfai, dfqi]).sort_values(by=sort_cols, ignore_index=True)
    df.columns.name = None
    return df


def adjust_unit(df: pd.DataFrame, unit: float) -> pd.DataFrame:
    currency_cols = [
        "total_assets",
        "current_assets",
        "current_liabilities",
        "equity",
        "revenues",
        "gross_profit",
        "ebit",
        "ebt",
        "effective_tax",
        "net_income",
        "operating_cash_flow",
        "depreciation_amortization",
        "total_cash",
        "total_debt",
        "net_debt",
        "working_capital",
        "ebitda",
        "invested_capital",
    ]
    df.loc[:, currency_cols] /= unit
    return df


def reorder_index(df: pd.DataFrame) -> pd.DataFrame:
    new_order = [
        "total_assets",
        "current_assets",
        "total_cash",
        "working_capital",
        "invested_capital",
        "current_liabilities",
        "total_debt",
        "net_debt",
        "equity",
        "revenues",
        "gross_profit",
        "net_income",
        "ebitda",
        "ebit",
        "ebt",
        "effective_tax",
        "operating_cash_flow",
        "depreciation_amortization",
        "effective_tax_rate",
        "return_on_assets",
        "return_on_equity",
        "roic",
        "gross_margin",
        "ebitda_margin",
        "operating_margin",
        "net_margin",
        "eps",
    ]
    return df.reindex(new_order)


def format_indicators(df: pd.DataFrame, unit: float) -> pd.DataFrame:
    df = adjust_unit(df, unit)
    melt_cols = ["cvm_id", "is_annual", "is_consolidated", "period_end"]
    df = pd.melt(df, id_vars=melt_cols, var_name="indicator", value_name="value")

    sort_cols = ["cvm_id", "is_consolidated", "period_end", "indicator"]
    df.sort_values(by=sort_cols, inplace=True)

    df["period_end"] = df["period_end"].astype("string")

    index_cols = ["cvm_id", "is_consolidated", "indicator"]
    df = (
        pd.pivot(df, values="value", index=index_cols, columns=["period_end"])
        .reset_index()
        .set_index("indicator")
    )
    df.columns.name = None
    df.index.name = None
    df = reorder_index(df)
    return df



"""Finlogic manager module.

This module provides upper level functions to handle financial data from the CVM
Portal. It allows updating, processing and consolidating financial statements,
as well as searching for company names in the FinLogic Database and retrieving
information about the database itself.
"""
from typing import Literal
import pandas as pd
from . import indicators as ind

CHECKMARK = "\033[32m\u2714\033[0m"
# Load last session data
TRADE_DATA_URL = (
    "https://raw.githubusercontent.com/crdcj/FinLogic/main/data/trades.csv.gz"
)
TRADED_FINANCIALS_URL = "https://raw.githubusercontent.com/crdcj/FinLogic/main/data/traded_companies_financials.csv.gz"
NOT_TRADED_FINANCIALS_URL = "https://raw.githubusercontent.com/crdcj/FinLogic/main/data/not_traded_companies_financials.csv.gz"
LANGUAGE_DATA_URL = (
    "https://raw.githubusercontent.com/crdcj/FinLogic/main/data/pten_df.csv.gz"
)
FINANCIALS_DF = pd.DataFrame()
TRADES_DF = pd.DataFrame()
LANGUAGE_DF = pd.DataFrame()
INDICATORS_DF = pd.DataFrame()


def load(is_traded: bool = True, min_volume: int = 100_000):
    """Verify changes in CVM files and update Finlogic Database if necessary.

    Args:
        is_traded (bool, optional): If True, only currently traded companies are
        loaded.
        min_volume (int): The minimum daily volume of the stock. Defaults
            to R$ 100k (aprox. USD 20k), which is a reasonable value to remove
            extremely illiquid stocks.

    Returns:
        None
    """
    global LANGUAGE_DF
    global TRADES_DF
    global FINANCIALS_DF
    print('Loading "language" data...')
    LANGUAGE_DF = pd.read_csv(LANGUAGE_DATA_URL)
    print("Loading trading data...")
    TRADES_DF = pd.read_csv(TRADE_DATA_URL)
    print("Loading financials data...")
    date_cols = ["period_begin", "period_end"]
    FINANCIALS_DF = pd.read_csv(TRADED_FINANCIALS_URL, parse_dates=date_cols)
    if not is_traded:
        FINANCIALS_DF = pd.concat(
            [FINANCIALS_DF, pd.read_csv(NOT_TRADED_FINANCIALS_URL)], ignore_index=True
        )
    TRADES_DF = TRADES_DF.query("volume >= @min_volume")
    TRADED_CVM_IDS = TRADES_DF["cvm_id"].unique()  # noqa
    FINANCIALS_DF = FINANCIALS_DF.query("cvm_id in @TRADED_CVM_IDS").reset_index(
        drop=True
    )
    print("Building indicators data...")
    global INDICATORS_DF
    INDICATORS_DF = ind.build_indicators(FINANCIALS_DF)
    print(f"{CHECKMARK} FinLogic is ready!")


def info() -> pd.DataFrame:
    """Print a concise summary of FinLogic available data.

    This function returns a dataframe containing main information about
    FinLogic Database, such as the database path, file size, last update call,
    last modified dates, size in memory, number of accounting rows, unique
    accounting codes, companies, unique financial statements, first financial
    statement date and last financial statement date.

    Args:
        return_dict (bool, optional): If True, returns a dictionary with the
            database information and do not print it.

    Returns: None
    """
    info = {}
    if FINANCIALS_DF.empty:
        return pd.DataFrame()

    info["data_url"] = f"{TRADED_FINANCIALS_URL}"
    data_size = (
        FINANCIALS_DF.memory_usage(deep=True).sum()
        + TRADES_DF.memory_usage(deep=True).sum()
    )
    info["memory_usage"] = f"{data_size / 1024**2:.1f} MB"
    # info["updated_on"] = db_last_modified.strftime("%Y-%m-%d %H:%M:%S")

    info["accounting_entries"] = FINANCIALS_DF.shape[0]

    report_cols = ["cvm_id", "is_annual", "period_end"]
    info["number_of_reports"] = FINANCIALS_DF[report_cols].drop_duplicates().shape[0]
    info["first_report"] = FINANCIALS_DF["period_end"].min().strftime("%Y-%m-%d")
    info["last_report"] = FINANCIALS_DF["period_end"].max().strftime("%Y-%m-%d")

    info["number_of_companies"] = FINANCIALS_DF["cvm_id"].nunique()

    s = pd.Series(info)
    s.name = "FinLogic Info"

    return s.to_frame()


def search_segment(search_value: str):
    series = TRADES_DF["segment"].drop_duplicates().sort_values(ignore_index=True)
    mask = series.str.contains(search_value)
    return series[mask].reset_index(drop=True)


def search_company(
    search_value: str,
    search_by: Literal["name_id", "cvm_id", "tax_id", "segment"] = "name_id",
) -> pd.DataFrame:
    """Search for a company name in FinLogic Database.

    This function searches the specified column in the FinLogic Database for
    company names that contain the provided expression. It returns a DataFrame
    containing the search results, with each row representing a unique company
    that matches the search criteria.

    Args:
        search_value (str): The search expression.
        search_by (str): The column where the id search will be performed. Valid
            values are 'name_id', 'cvm_id', and 'tax_id'. Defaults to 'name_id'.

    Returns:
        pd.DataFrame: A DataFrame containing the search results, with columns
            'name_id', 'cvm_id', and 'tax_id' for each unique company that
            matches the search criteria.
    """
    search_cols = ["name_id", "cvm_id", "tax_id"]
    df = FINANCIALS_DF[search_cols].drop_duplicates(
        subset=["cvm_id"], ignore_index=True
    )
    df = pd.merge(df, TRADES_DF, on="cvm_id")
    match search_by:
        case "name_id":
            # Company name is stored in uppercase in the database
            df.query(f"name_id.str.contains('{search_value.upper()}')", inplace=True)
        case "cvm_id":
            df.query(f"cvm_id == {search_value}", inplace=True)
        case "tax_id":
            df.query(f"tax_id == '{search_value}'", inplace=True)
        case "segment":
            df.query(f"segment.str.contains('{search_value}')", inplace=True)
        case _:
            raise ValueError("Invalid value for 'search_by' argument.")

    show_cols = [
        "name_id",
        "cvm_id",
        "tax_id",
        "segment",
        "is_restructuring",
        "most_traded_stock",
    ]
    return df[show_cols].reset_index(drop=True)


def rank(
    segment: str = None,
    n: int = 10,
    rank_by: str = "operating_margin",
    is_consolidated: bool = True,
) -> pd.DataFrame:
    """Rank companies by a given indicator.

    This function returns a DataFrame containing the top n companies in the
    specified segment, ranked by the given indicator.

    Args:
        segment (str): The segment to be ranked. Defaults to None, which
            returns the top n companies in all segments.
        n (int): The number of companies to be returned. Defaults to 10.
        rank_by (str): The indicator to be used for ranking. Defaults to
            'operating_margin'. Valid values are 'total_assets', 'equity',
            'revenues', 'gross_profit', 'ebit', 'ebt', 'net_income',
            'operating_cash_flow', 'eps', 'total_cash', 'total_debt',
            'net_debt', 'ebitda', 'gross_margin', 'ebitda_margin',
            'operating_margin', 'net_margin', 'return_on_assets',
            'return_on_equity', 'roic'.
    """
    show_cols = [
        "name_id",
        "most_traded_stock",
        "cvm_id",
        "is_restructuring",
        "is_consolidated",
        "segment",
        "period_end",
        rank_by,
    ]
    df = (
        FINANCIALS_DF.sort_values(
            by=["cvm_id", "period_end", "is_consolidated"], ignore_index=True
        )
        .drop_duplicates(subset=["cvm_id"], keep="last")
        .merge(TRADES_DF, on="cvm_id")
        .merge(
            INDICATORS_DF[["cvm_id", rank_by, "is_consolidated", "period_end"]],
            on=["cvm_id", "period_end", "is_consolidated"],
        )
        .query("segment.str.contains(@segment) and is_consolidated == @is_consolidated")
        .sort_values(by=[rank_by], ascending=False, ignore_index=True)
        .head(n)[show_cols]
    )

    return df
